package net.sourceforge.chopchophttpclient.processor;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetSocketAddress;
import java.net.Proxy;
import java.net.ProxySelector;
import java.net.URI;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import net.sourceforge.chopchophttpclient.Client;
import net.sourceforge.chopchophttpclient.CollectiveProgressListener;
import net.sourceforge.chopchophttpclient.DownloadResult;
import net.sourceforge.chopchophttpclient.PostParameter;
import net.sourceforge.chopchophttpclient.ProgressListener;
import net.sourceforge.chopchophttpclient.util.RetrievableThread;

import org.apache.commons.httpclient.Header;
import org.apache.commons.httpclient.HostConfiguration;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpException;
import org.apache.commons.httpclient.HttpMethod;
import org.apache.commons.httpclient.MultiThreadedHttpConnectionManager;
import org.apache.commons.httpclient.NameValuePair;
import org.apache.commons.httpclient.StatusLine;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.commons.httpclient.methods.HeadMethod;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.commons.httpclient.util.DateParseException;
import org.apache.commons.httpclient.util.DateUtil;
import org.apache.log4j.Logger;


/**
 * <p style="font-weight: bold;">Note:</p> HTTP 301,302,307 redirects are followed
 * 
 * <p style="font-weight: bold;">Note:</p> Thread safe
 * 
 * @author rbeede
 *
 */
public class DownloadRunnable implements Runnable {
	private static final Logger log = Logger.getLogger(DownloadRunnable.class);
	
	private final HttpClient httpClient;
	private final URI uri;
	private final OutputStream outputStream;
	private final PostParameter[] postParameters;
	private final ProxySelector proxySelector;
	private final ProgressListener progressListener;
	
	private DownloadResult downloadResult;
	
	private final HostConfiguration hostConfiguration;

	
	/**
	 * @param httpClient Make sure you are using {@link MultiThreadedHttpConnectionManager}.  May not be null.
	 * @param uri Absolute URI to access.  May not be null.
	 * @param outputStream Where to write the content.  May not be null.
	 * @param postParameters see {@link Client#download(URI[], java.io.File[], PostParameter[])}.  May be null.
	 * @param proxySelector see {@link Client#Client(String, String, ProxySelector, ProgressListener)}.  May be null.
	 * @param progressListener see {@link Client#Client(String, String, ProxySelector, ProgressListener)}.  May be null.
	 * @throws NullPointerException if any non-allowed arguments are null
	 */
	public DownloadRunnable(HttpClient httpClient, URI uri, OutputStream outputStream,
			PostParameter[] postParameters, ProxySelector proxySelector, ProgressListener progressListener) {
		super();
		
		if(null == httpClient) {
			throw new NullPointerException("httpClient cannot be null");
		}
		if(null == uri) {
			throw new NullPointerException("uri cannot be null");
		}
		if(null == outputStream) {
			throw new NullPointerException("outputStream cannot be null");
		}
		
		this.httpClient = httpClient;
		this.uri = uri;
		this.outputStream = outputStream;
		if(null != postParameters) {
			//Copy array so external changes to array (like sorting) don't affect our array
			this.postParameters = net.sourceforge.chopchophttpclient.util.Arrays.copyOf(postParameters, postParameters.length);
		} else {
			this.postParameters = null;
		}
		this.proxySelector = proxySelector;
		this.progressListener = progressListener;
		
		
		this.hostConfiguration = new HostConfiguration();
		/* Grab a suitable proxy (if desired) */
		final Proxy proxy = getProxy(this.proxySelector, this.uri);
		
		if(!Proxy.NO_PROXY.equals(proxy)) {
			//We need a proxy so configure it
			
			final InetSocketAddress proxyAddress = (InetSocketAddress) proxy.address();
			final String proxyHost = proxyAddress.getAddress().getHostAddress();
			final int proxyPort = proxyAddress.getPort();
			
			if(Proxy.Type.SOCKS.equals(proxy.type())) {
				//We must set a special property that applies to the entire JVM
				System.setProperty("socksProxyHost", proxyHost);
				System.setProperty("socksProxyPort", Integer.toString(proxyPort));
				
				// Username and password happen care of 
				//	http://java.sun.com/javase/6/docs/technotes/guides/net/properties.html
				// This should have already been done for this.httpClient which was passed to this class
			} else {
				hostConfiguration.setProxy(proxyHost, proxyPort);
			}
		}
		
		/* Login credentials (remote site and proxy) should already have been setup in this.httpClient */
	}


	public void run() {
		final Date startDate = new Date();

		/* At this point see if the remote server supports accelerated downloads via breaking a download into
		 * multiple pieces care of the Accept-Ranges HTTP header 
		*/
		final long acceptsRanges = useRemoteAcceptRanges();  //If positive then yes and represents content length
		
		
		if(acceptsRanges > 0) {
			downloadRanges(startDate, acceptsRanges);  // Will update this.downloadResult
		} else {
			//Just pull it down in one single download
			downloadNoRanges(startDate);  // Will update this.downloadResult
		}
	}
	
	
	/**
	 * <p style="font-weight: bold;">Note:</p> Does not configure the proxy for use.  Simply tells what would be
	 * 		appropriate to use.
	 * 
	 * <p style="font-weight: bold;">Note:</p> In the event of multiple proxies returns the first proxy given
	 * 
	 * @param proxySelector Selector for getting back a list of valid proxies
	 * @param absoluteURI Absolute URI to compare for an appropriate proxy
	 * @return Proxy if required.  Proxy.NO_PROXY otherwise.
	 */
	private static Proxy getProxy(final ProxySelector proxySelector, final URI absoluteURI) {
		if(null == proxySelector) {
			return Proxy.NO_PROXY;
		}

		final List<Proxy> proxies = proxySelector.select(absoluteURI);
		
		if(null == proxies || proxies.isEmpty()) {
			return Proxy.NO_PROXY;
		} else {
			return proxies.get(0);
		}
	}
	
	
	/**
	 * @return null if never started or still running.  DownloadResult otherwise.
	 */
	public DownloadResult getDownloadResult() {
		return this.downloadResult;
	}

	
	protected static String getUserAgent() {
		final String[] systemProperites = {"java.version", "java.vendor", "os.name", "os.arch", "os.version"};
		
		final StringBuilder sb = new StringBuilder(255);
		
		sb.append(DownloadRunnable.class.getPackage().getName());
		sb.append('/');
		
		sb.append("ImplementationVersion=");
		sb.append(Client.class.getPackage().getImplementationVersion());
		sb.append('/');

		for(final String property : systemProperites) {
			sb.append(property);
			sb.append('=');
			sb.append(System.getProperty(property));
			sb.append('/');
		}


		return sb.toString();
	}
	
	
	/**
	 * @see <a href="http://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.3.1">"http://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.3.1"</a>
	 * 
	 * @param httpDate The date according to RFC2616
	 * @return null if date could not be parsed.  Parsed date otherwise.
	 */
	protected static java.util.Date parseHttpDate(final String httpDate) {
		if(null == httpDate)  return null;
		
		//HTTP/1.1 protocol (RFC2616) requires 3 possible date formats
		//We'll use Apache's Jakarta HTTPClient to parse the dates for us
		java.util.Date parsedDate = null;
		try {
			 parsedDate = DateUtil.parseDate(httpDate);
		} catch (DateParseException e) {
			return null;
		}
		
		return parsedDate;
	}

	
	/**
	 * Records a failure with as much information as was available
	 * 
	 * <p style="font-weight: bold;">Post-conditions:</p>
	 * <ul>
	 * 	<li>Always closes {@link #outputStream}</li>
	 * 	<li>Always calls {@link HttpMethod#releaseConnection()}</li>
	 * 	<li>Sets {@link #downloadResult} with failure information</li>
	 * </ul>
	 * 
	 * @param method The HTTP method
	 * @param t The error
	 */
	private void recordFailureResult(final HttpMethod method, final Throwable t, final Date startDate) {
		String statusText = null;
		int statusCode = Integer.MIN_VALUE;
		
		
		if(null != method && method.hasBeenUsed()) {
			StatusLine statusLine;
			try {
				statusLine = method.getStatusLine();
			} catch(NullPointerException npe) {
				//Jakarta HTTPClient 3.1 has a tendency to throw a NPE at times based on the incomplete method state
				log.trace(npe,npe);
				statusLine = null;
			}
			
			if(null != statusLine) {
				statusText = statusLine.getReasonPhrase();
				statusCode = statusLine.getStatusCode();
			}
		}

		recordFailureResult(statusText, statusCode, t, startDate);
		
		if(null != method) {
			method.releaseConnection();
		}
	}
	
	
	private void recordFailureResult(final String statusText, final int statusCode, final Throwable t, final Date startDate) {
		this.downloadResult = new DownloadResult(
				this.uri,
				statusCode,  // may be null
				statusText,  // may be negative
				Long.MIN_VALUE,
				Long.MIN_VALUE,
				Long.MIN_VALUE,
				t,
				startDate.getTime(),
				0,  //no end time
				this.outputStream
				);

		try {
			this.outputStream.close();
		} catch(Exception npExcep) {
			;  //Quietly done
		}
	}
	
	
	/**
	 * @return Content length in bytes if remote server accepts ranges AND the content
	 * length >= {@link Client#MIN_SIZE_USE_PIECES}.  Negative if server does not or could not be reached.
	 * 
	 * @see HTTP 1.1 specification section 14.5 of RFC2616
	 */
	private long useRemoteAcceptRanges() {
		final HeadMethod method = new HeadMethod(this.uri.toString());
		
		//Common options
		method.setRequestHeader("User-Agent", getUserAgent());

		try {
			this.httpClient.executeMethod(this.hostConfiguration, method);
		} catch (HttpException e) {
			return Long.MIN_VALUE;
		} catch (IOException e) {
			return Long.MIN_VALUE;
		}
		
		final Header acceptRangesHeader = method.getResponseHeader("Accept-Ranges");
		
		if(null == acceptRangesHeader || !"bytes".equalsIgnoreCase(acceptRangesHeader.getValue())) {
			return Long.MIN_VALUE;
		}
		
		
		final Header contentLengthHeader = method.getResponseHeader("Content-Length");
		
		if(null == contentLengthHeader) {
			return Long.MIN_VALUE;
		}
		
		final long contentLength;
		try {
			contentLength = Long.parseLong(contentLengthHeader.getValue());
		} catch(NumberFormatException e) {
			log.trace("Content-Length header had non-parsable value of " + contentLengthHeader.getValue(),e);
			return Long.MIN_VALUE;
		}
		
		if(contentLength < Client.MIN_SIZE_USE_PIECES) {
			return Long.MIN_VALUE;
		}
		
		
		return contentLength;  // Must be okay
	}
	
	
	/**
	 * <p style="font-weight: bold;">Post-conditions</p>
	 * <ul>
	 * 	<li>Modifies {@link #downloadResult} with appropriate information</li>
	 * </ul>
	 * 
	 * @param startDate Timestamp to record in {@link #downloadResult}
	 */
	private void downloadNoRanges(final Date startDate) {
		final HttpMethod method;
		
		/* Setup our HTTP method */
		if(null == this.postParameters) {
			//relative, will be resolved against base in httpClient
			method = new GetMethod(this.uri.toString());
			method.setFollowRedirects(true);  //Only allowed for GET, not possible with POST
		} else {
			method = new PostMethod(this.uri.toString());
			
			final NameValuePair[] data = new NameValuePair[this.postParameters.length];
			for(int i = 0; i < this.postParameters.length; i++) {
				data[i] = new NameValuePair(
						this.postParameters[i].getName(),
						this.postParameters[i].getValue()
						);
			}
			
			((PostMethod) method).setRequestBody(data);
		}
		
		//Common options
		method.setRequestHeader("User-Agent", DownloadRunnable.getUserAgent());

		
		/* Execute the method! */
		try {
			this.httpClient.executeMethod(this.hostConfiguration, method);
		} catch (HttpException e) {
			recordFailureResult(method, e, startDate);
			return;
		} catch (IOException e) {
			recordFailureResult(method, e, startDate);
			return;
		}
		
		
		/* At this point we got some kind of response back (success or error)
		 * 
		 * So we go on and read whatever is available if anything for the DownloadResult
		 * 
		 * Even if the remote server returned an error we can read the content for any error messages
		 */
		
		
		/* Content Length */
		final long remoteContentLength;  //As reported by remote (may be before uncompressed locally)
		final Header contentLengthHeader = method.getResponseHeader("Content-Length");
		if(null != contentLengthHeader) {
			try {
				remoteContentLength = Long.parseLong(contentLengthHeader.getValue());
			} catch(NumberFormatException e) {
				final Exception excep = new Exception("Invalid Content-Length of '" + contentLengthHeader.getValue() + "'", e);
				
				recordFailureResult(method, excep, startDate);
				return;
			}
		} else {
			log.trace("No Content-Length returned for " + this.uri.toString());
			remoteContentLength = -1;  //Unknown length
		}
		
		
		/* Read the actual content */
		final InputStream responseStream;
		
		try {
			responseStream = method.getResponseBodyAsStream();
		} catch(IOException e) {
			recordFailureResult(method, e, startDate);
			return;
		}
		
		//We must track how much we've read since the connection is probably a keep-alive one
		int totalBytesRead = 0;
		int lastNumBytesRead = 0;  //bytes just read from last read
		byte[] buffer = new byte[Client.BUFFER_SIZE];
		
		try {
			while(totalBytesRead != remoteContentLength && -1 != (lastNumBytesRead = responseStream.read(buffer))) {
				this.outputStream.write(buffer, 0, lastNumBytesRead);
				
				totalBytesRead += lastNumBytesRead;
				
				if(null != this.progressListener) {
					this.progressListener.update(this.uri, totalBytesRead, remoteContentLength);
				}
			}
		} catch (IOException e) {
			recordFailureResult(method, e, startDate);
			return;
		}
		
		if(-1 != remoteContentLength && remoteContentLength != totalBytesRead) {
			//Didn't download correct amount of known content length bytes
			HttpException e = new HttpException("Total bytes read was " + totalBytesRead + " but expected " + remoteContentLength);
			
			recordFailureResult(method, e, startDate);
			return;
		}

		
		try {
			this.outputStream.close();
		} catch(IOException e) {
			//Stream isn't in state that needs closed
			//quietly go on
		}
		
		
		/* Grab the last modified date if available */
		final Header lastModified = method.getResponseHeader("Last-Modified");
		long lastModifiedEpoch = Long.MIN_VALUE;
		if(null != lastModified && null != parseHttpDate(lastModified.getValue())) {
			lastModifiedEpoch = parseHttpDate(lastModified.getValue()).getTime();
		}
		
		/* Store the successful download result */
		this.downloadResult = new DownloadResult(
				this.uri,
				method.getStatusCode(),
				method.getStatusText(),
				remoteContentLength,
				totalBytesRead,
				lastModifiedEpoch,
				null,  //no Exception
				startDate.getTime(),
				(new Date()).getTime(),  //end time
				this.outputStream
				);

		method.releaseConnection();  //We're all done with the method data now
	}


	/**
	 * <p style="font-weight: bold;">Post-conditions</p>
	 * <ul>
	 * 	<li>Modifies {@link #downloadResult} with appropriate information</li>
	 * </ul>
	 * 
	 * @param startDate Timestamp to record in {@link #downloadResult}
	 */
	private void downloadRanges(final Date startDate, final long contentLength) {
		log.trace("Using " + Client.DOWNLOAD_PIECES + " range connections for " + this.uri.toString());
		
		
		final List<RetrievableThread<PieceDownloadRunnable>> piecesThreads = new ArrayList<RetrievableThread<PieceDownloadRunnable>>(Client.DOWNLOAD_PIECES);
		
		
		// We need to setup a special ProgressListener wrapper
		final CollectiveProgressListener collectivePL = new CollectiveProgressListener(this.progressListener);
		
		
		// Now setup each PieceDownloadRunnable with Thread
		final long maxPieceSizes = (long) Math.ceil((contentLength / Client.DOWNLOAD_PIECES));
		for(int i = 0; i < Client.DOWNLOAD_PIECES; i++) {
			final long startByte = maxPieceSizes * i;
			long endByte = (maxPieceSizes * (i+1)) - 1;
			
			if(endByte > contentLength) {
				endByte = contentLength;
			}
			
			RetrievableThread<PieceDownloadRunnable> thread;
			try {
				thread = new RetrievableThread<PieceDownloadRunnable>(
						new PieceDownloadRunnable(this.httpClient, this.uri, this.postParameters, this.hostConfiguration, collectivePL.getMember(), startByte, endByte)
					);
			} catch (IOException e) {
				recordFailureResult(null, e, startDate);
				return;
			}
			
			thread.start();  // Fire it off
			
			piecesThreads.add(thread);
			
			log.trace("Started thread number " + (i+1) + " for piece download of " + this.uri.toString());
		}
		
		
		boolean allDone;
		do {
			allDone = true;
			
			for(RetrievableThread<PieceDownloadRunnable> thread : piecesThreads) {
				if(!Thread.State.TERMINATED.equals(thread.getState())) {
					allDone = false;
				}
			}
			
			Thread.yield();
		} while(!allDone);

		
		// All piece threads have completed (successful or not)
		// Check for any errors and reassemble the data
		for(RetrievableThread<PieceDownloadRunnable> thread : piecesThreads) {
			final DownloadResult pieceDownloadResult = thread.getRunnable().getDownloadResult();
			
			if(null != pieceDownloadResult.getError()) {
				// Touch luck
				recordFailureResult(pieceDownloadResult.getServerResponseString(),
									pieceDownloadResult.getServerResponseCode(),
									pieceDownloadResult.getError(),
									startDate);
				return;
			} else if(pieceDownloadResult.getServerResponseCode() < 200 || pieceDownloadResult.getServerResponseCode() > 299) {
				recordFailureResult(pieceDownloadResult.getServerResponseString(),
						pieceDownloadResult.getServerResponseCode(),
						pieceDownloadResult.getError(),
						startDate);
				return;
			}
			
			
			// Retrieve the content which we known is a ByteArrayOutputStream
			final ByteArrayOutputStream baosContent = (ByteArrayOutputStream) pieceDownloadResult.getOutputStreamDestination();
			
			// Dump the data into the given OutputStream (adding this piece to the end)
			try {
				baosContent.writeTo(this.outputStream);
			} catch (IOException e) {
				recordFailureResult(pieceDownloadResult.getServerResponseString(),
						pieceDownloadResult.getServerResponseCode(),
						e,
						startDate);
			}
		}
		
		
		// Now cleanly close
		try {
			this.outputStream.close();
		} catch(IOException e) {
			//Stream isn't in state that needs closed
			//quietly go on
		}
		
		
		/* Store the successful download result (using first piece meta data info) */
		final DownloadResult firstDownloadResult = piecesThreads.get(0).getRunnable().getDownloadResult();
		
		this.downloadResult = new DownloadResult(
				this.uri,
				firstDownloadResult.getServerResponseCode(),
				firstDownloadResult.getServerResponseString(),
				contentLength,
				contentLength,  // total bytes should be this since everything went okay
				firstDownloadResult.getLastModified(),
				null,  //no Exception
				startDate.getTime(),
				(new Date()).getTime(),  //end time
				this.outputStream
				);
		
	}
}
