package net.sourceforge.chopchophttpclient.processor;

import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.util.Date;

import net.sourceforge.chopchophttpclient.Client;
import net.sourceforge.chopchophttpclient.DownloadResult;
import net.sourceforge.chopchophttpclient.PostParameter;
import net.sourceforge.chopchophttpclient.ProgressListener;
import net.sourceforge.chopchophttpclient.util.MemoryAndDiskOutputStream;

import org.apache.commons.httpclient.Header;
import org.apache.commons.httpclient.HostConfiguration;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpException;
import org.apache.commons.httpclient.HttpMethod;
import org.apache.commons.httpclient.NameValuePair;
import org.apache.commons.httpclient.StatusLine;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.log4j.Logger;

public class PieceDownloadRunnable implements Runnable {
	private static final Logger log = Logger.getLogger(PieceDownloadRunnable.class);
	
	public static final int IN_MEMORY_BUFFER = 2 * 1024 * 1024;  //2 MiB
	
	private final HttpClient httpClient;
	private final PostParameter[] postParameters;
	private final URI uri;
	private final HostConfiguration hostConfiguration;
	private final ProgressListener progressListener;
	
	private DownloadResult downloadResult;  //Will have OutputStream with piece data
	
	// These are inclusive and zero indexed
	//	See RFC2616	14.35.1
	private final long startByte;
	private final long endByte;
	
	private final long pieceContentLength;
	
	private final MemoryAndDiskOutputStream outputStream;
	
	/**
	 * @param httpClient
	 * @param uri
	 * @param postParameters
	 * @param hostConfiguration
	 * @param progressListener
	 * @param startByte
	 * @param endByte
	 * @throws IOException If an error occurred setting up temporary files on disk for storage
	 */
	public PieceDownloadRunnable(HttpClient httpClient, URI uri, PostParameter[] postParameters, 
			HostConfiguration hostConfiguration, ProgressListener progressListener, long startByte, long endByte) throws IOException {
		
		this.httpClient = httpClient;
		this.uri = uri;
		this.postParameters = postParameters;
		this.hostConfiguration = hostConfiguration;
		this.progressListener = progressListener;
		
		this.startByte = startByte;
		this.endByte = endByte;
		
		this.pieceContentLength = (int) (this.endByte - this.startByte) + 1;
		
	
		try {
			this.outputStream = new MemoryAndDiskOutputStream(IN_MEMORY_BUFFER);
		} catch(IOException e) {
			log.fatal("Unable to create temporary files on this system.", e);
			throw e;
		}
	}
	
	public void run() {
		final Date startDate = new Date();
				
		
		// Did we exceed the maximum byte size we can currently handle?
		if(	(this.endByte - this.startByte) > Integer.MAX_VALUE	) {
			recordFailureResult(null, new IndexOutOfBoundsException("Piece data size of " + (this.endByte - this.startByte) + " is to large"), startDate);
			return;
		}
		
		
		final HttpMethod method;
		
		/* Setup our HTTP method */
		if(null == this.postParameters) {
			//relative, will be resolved against base in httpClient
			method = new GetMethod(this.uri.toString());
			method.setFollowRedirects(true);  //Only allowed for GET, not possible with POST
		} else {
			method = new PostMethod(this.uri.toString());
			
			final NameValuePair[] data = new NameValuePair[this.postParameters.length];
			for(int i = 0; i < this.postParameters.length; i++) {
				data[i] = new NameValuePair(
						this.postParameters[i].getName(),
						this.postParameters[i].getValue()
						);
			}
			
			((PostMethod) method).setRequestBody(data);
		}
		
		//Common options
		method.setRequestHeader("User-Agent", DownloadRunnable.getUserAgent());
		
		
		//Our range option
		method.setRequestHeader("Range", "bytes=" + this.startByte + "-" + this.endByte);

		
		/* Execute the method! */
		try {
			this.httpClient.executeMethod(this.hostConfiguration, method);
		} catch (HttpException e) {
			recordFailureResult(method, e, startDate);
			return;
		} catch (IOException e) {
			recordFailureResult(method, e, startDate);
			return;
		}
	

		/* At this point we got some kind of response back (success or error)
		 * 
		 * So we go on and read whatever is available if anything for the DownloadResult
		 * 
		 * Even if the remote server returned an error we can read the content for any error messages
		 */
		
		
		/* Read the actual content */
		final InputStream responseStream;
		
		try {
			responseStream = method.getResponseBodyAsStream();
		} catch(IOException e) {
			recordFailureResult(method, e, startDate);
			return;
		}

		
		//We must track how much we've read since the connection is probably a keep-alive one
		int totalBytesRead = 0;
		int lastNumBytesRead = 0;  //bytes just read from last read
		byte[] buffer = new byte[Client.BUFFER_SIZE];
		
		try {
			while(totalBytesRead != this.pieceContentLength && -1 != (lastNumBytesRead = responseStream.read(buffer))) {
				try {
					this.outputStream.write(buffer, 0, lastNumBytesRead);
				} catch(java.lang.OutOfMemoryError e) {
					final java.lang.OutOfMemoryError detailedExcep = new java.lang.OutOfMemoryError("Tried to write " + lastNumBytesRead + " bytes");
					detailedExcep.initCause(e);
					
					log.fatal(detailedExcep, detailedExcep);
					throw detailedExcep;
				}
				
				totalBytesRead += lastNumBytesRead;
				
				if(totalBytesRead > pieceContentLength) {
					// Server gave back too much content
					HttpException e = new HttpException("Aborted download.  Total bytes read was " + totalBytesRead + " but expected piece length of " + this.pieceContentLength);
					
					recordFailureResult(method, e, startDate);
					return;
				}
				
				if(null != this.progressListener) {
					this.progressListener.update(this.uri, totalBytesRead, pieceContentLength);
				}
			}
		} catch (IOException e) {
			recordFailureResult(method, e, startDate);
			return;
		}
		
		if(-1 != this.pieceContentLength && this.pieceContentLength != totalBytesRead) {
			//Didn't download correct amount of known content length bytes
			HttpException e = new HttpException("Total bytes read was " + totalBytesRead + " but expected piece length of " + this.pieceContentLength);
			
			recordFailureResult(method, e, startDate);
			return;
		}

		
		try {
			// No effect for ByteArrayOutputStream today, but good if we change it in the future
			this.outputStream.close();
		} catch(IOException e) {
			//Stream isn't in state that needs closed
			//quietly go on
		}
		
		
		/* Grab the last modified date if available */
		final Header lastModified = method.getResponseHeader("Last-Modified");
		long lastModifiedEpoch = Long.MIN_VALUE;
		if(null != lastModified && null != DownloadRunnable.parseHttpDate(lastModified.getValue())) {
			lastModifiedEpoch = DownloadRunnable.parseHttpDate(lastModified.getValue()).getTime();
		}
		
		/* Store the successful download result */
		this.downloadResult = new DownloadResult(
				this.uri,
				method.getStatusCode(),
				method.getStatusText(),
				this.pieceContentLength,
				totalBytesRead,
				lastModifiedEpoch,
				null,  //no Exception
				startDate.getTime(),
				(new Date()).getTime(),  //end time
				this.outputStream
				);

		method.releaseConnection();  //We're all done with the method data now
	}
	
	
	/**
	 * Records a failure with as much information as was available
	 * 
	 * <p style="font-weight: bold;">Post-conditions:</p>
	 * <ul>
	 * 	<li>Always closes {@link #outputStream}</li>
	 * 	<li>Always calls {@link HttpMethod#releaseConnection()}</li>
	 * 	<li>Sets {@link #downloadResult} with failure information</li>
	 * </ul>
	 * 
	 * @param method The HTTP method
	 * @param t The error
	 */
	private void recordFailureResult(final HttpMethod method, final Exception e, final Date startDate) {
		String statusText = null;
		int statusCode = Integer.MIN_VALUE;
		
		
		if(null != method && method.hasBeenUsed()) {
			StatusLine statusLine;
			try {
				statusLine = method.getStatusLine();
			} catch(NullPointerException npe) {
				//Jakarta HTTPClient 3.1 has a tendency to throw a NPE at times based on the incomplete method state
				log.trace(npe,npe);
				statusLine = null;
			}
			
			if(null != statusLine) {
				statusText = statusLine.getReasonPhrase();
				statusCode = statusLine.getStatusCode();
			}
		}
		
		
		this.downloadResult = new DownloadResult(
				this.uri,
				statusCode,
				statusText,
				Long.MIN_VALUE,
				Long.MIN_VALUE,
				Long.MIN_VALUE,
				e,
				startDate.getTime(),
				0,  //no end time
				this.outputStream
				);
		

		try {
			this.outputStream.close();
		} catch(Exception npExcep) {
			;  //Quietly done
		}
		
		if(null != method) {
			method.releaseConnection();
		}
	}
	
	
	/**
	 * @return null if never started or still running.  DownloadResult otherwise.
	 */
	public DownloadResult getDownloadResult() {
		return this.downloadResult;
	}
}
